@@ -11,13 +11,15 @@ module Database.Persist.Postgresql
     , module Database.Persist.GenericSql
     , ConnectionString
     , PostgresConf (..)
+    , openSimpleConn
     ) where
 
 import Database.Persist hiding (Entity (..))
 import Database.Persist.Store
-import Database.Persist.GenericSql hiding (Key(..))
+import Database.Persist.GenericSql hiding (Key)
 import Database.Persist.GenericSql.Internal
 import Database.Persist.EntityDef
+import Data.Maybe (mapMaybe)
 
 import qualified Database.PostgreSQL.Simple as PG
 import qualified Database.PostgreSQL.Simple.BuiltinTypes as PG
@@ -31,7 +33,6 @@ import qualified Database.PostgreSQL.LibPQ as LibPQ
 
 import Control.Exception (throw)
 import Control.Monad.IO.Class (MonadIO (..))
-import Control.Monad.Trans.Control (MonadBaseControl)
 import Data.List (intercalate)
 import Data.IORef
 import qualified Data.Map as Map
@@ -39,18 +40,19 @@ import Data.Either (partitionEithers)
 import Control.Arrow
 import Data.List (sort, groupBy)
 import Data.Function (on)
-import qualified Data.Conduit as C
+import Data.Conduit
 import qualified Data.Conduit.List as CL
 
 import Data.ByteString (ByteString)
 import qualified Data.ByteString.Char8 as B8
 import qualified Data.Text as T
 import qualified Data.Text.Encoding as T
--- import Data.Time.LocalTime (localTimeToUTC, utc)
+import Data.Time.LocalTime (localTimeToUTC, utc)
 import Data.Text (Text, pack)
 import Data.Aeson
 import Control.Monad (forM, mzero)
 import System.Environment (getEnvironment)
+import Data.Int (Int64)
 
 
 -- | A @libpq@ connection string.  A simple example of connection
@@ -102,7 +104,11 @@ withPostgresqlConn = withSqlConn . open'
 
 open' :: ConnectionString -> IO Connection
 open' cstr = do
-    conn <- PG.connectPostgreSQL cstr
+    PG.connectPostgreSQL cstr >>= openSimpleConn
+
+-- | Generate a 'Connection' from a 'PG.Connection'
+openSimpleConn :: PG.Connection -> IO Connection
+openSimpleConn conn = do
     smap <- newIORef $ Map.empty
     return Connection
         { prepare    = prepare' conn
@@ -127,30 +133,28 @@ prepare' conn sql = do
         , withStmt = withStmt' conn query
         }
 
-insertSql' :: DBName -> [DBName] -> Either Text (Text, Text)
-insertSql' t cols = Left $ pack $ concat
+insertSql' :: DBName -> [DBName] -> DBName -> InsertSqlResult
+insertSql' t cols id' = ISRSingle $ pack $ concat
     [ "INSERT INTO "
     , T.unpack $ escape t
     , "("
     , intercalate "," $ map (T.unpack . escape) cols
     , ") VALUES("
     , intercalate "," (map (const "?") cols)
-    , ") RETURNING id"
+    , ") RETURNING "
+    , T.unpack $ escape id'
     ]
 
-execute' :: PG.Connection -> PG.Query -> [PersistValue] -> IO ()
-execute' conn query vals = do
-    _ <- PG.execute conn query (map P vals)
-    return ()
+execute' :: PG.Connection -> PG.Query -> [PersistValue] -> IO Int64
+execute' conn query vals = PG.execute conn query (map P vals)
 
-withStmt' :: C.MonadResource m
+withStmt' :: MonadResource m
           => PG.Connection
           -> PG.Query
           -> [PersistValue]
-          -> C.Source m [PersistValue]
-withStmt' conn query vals = C.sourceIO (liftIO   openS )
-                                       (liftIO . closeS)
-                                       (liftIO . pullS )
+          -> Source m [PersistValue]
+withStmt' conn query vals =
+    bracketP openS closeS pull
   where
     openS = do
       -- Construct raw query
@@ -173,8 +177,9 @@ withStmt' conn query vals = C.sourceIO (liftIO   openS )
                   LibPQ.TuplesOk -> return ()
                   _ -> do
                     msg <- LibPQ.resStatus status
+                    mmsg <- LibPQ.resultErrorMessage ret
                     fail $ "Postgresql.withStmt': bad result status " ++
-                           show status ++ " (" ++ show msg ++ ")"
+                           show status ++ " (" ++ (maybe (show msg) (show . (,) msg) mmsg) ++ ")"
 
                 -- Get number and type of columns
                 cols <- LibPQ.nfields ret
@@ -195,11 +200,17 @@ withStmt' conn query vals = C.sourceIO (liftIO   openS )
 
     closeS (ret, _, _, _) = LibPQ.unsafeFreeResult ret
 
+    pull x = do
+        y <- liftIO $ pullS x
+        case y of
+            Nothing -> return ()
+            Just z -> yield z >> pull x
+
     pullS (ret, rowRef, rowCount, getters) = do
         row <- atomicModifyIORef rowRef (\r -> (r+1, r))
         if row == rowCount
-           then return C.IOClosed
-           else fmap C.IOOpen $ forM (zip getters [0..]) $ \(getter, col) -> do
+           then return Nothing
+           else fmap Just $ forM (zip getters [0..]) $ \(getter, col) -> do
                                 mbs <- LibPQ.getvalue' ret row col
                                 case mbs of
                                   Nothing -> return PersistNull
@@ -220,6 +231,7 @@ instance PGTF.ToField P where
     toField (P (PersistDay d))         = PGTF.toField d
     toField (P (PersistTimeOfDay t))   = PGTF.toField t
     toField (P (PersistUTCTime t))     = PGTF.toField t
+    toField (P (PersistZonedTime (ZT t))) = PGTF.toField t
     toField (P PersistNull)            = PGTF.toField PG.Null
     toField (P (PersistList l))        = PGTF.toField $ listToJSON l
     toField (P (PersistMap m))         = PGTF.toField $ mapToJSON m
@@ -233,29 +245,30 @@ convertPV f = (fmap f .) . PGFF.fromField
 
 -- FIXME: check if those are correct and complete.
 getGetter :: PG.BuiltinType -> Getter PersistValue
-getGetter PG.Bool      = convertPV PersistBool
-getGetter PG.Bytea     = convertPV (PersistByteString . unBinary)
-getGetter PG.Char      = convertPV PersistText
-getGetter PG.Name      = convertPV PersistText
-getGetter PG.Int8      = convertPV PersistInt64
-getGetter PG.Int2      = convertPV PersistInt64
-getGetter PG.Int4      = convertPV PersistInt64
-getGetter PG.Text      = convertPV PersistText
-getGetter PG.Xml       = convertPV PersistText
-getGetter PG.Float4    = convertPV PersistDouble
-getGetter PG.Float8    = convertPV PersistDouble
-getGetter PG.Abstime   = convertPV PersistUTCTime
-getGetter PG.Reltime   = convertPV PersistUTCTime
-getGetter PG.Money     = convertPV PersistDouble
-getGetter PG.Bpchar    = convertPV PersistText
-getGetter PG.Varchar   = convertPV PersistText
-getGetter PG.Date      = convertPV PersistDay
-getGetter PG.Time      = convertPV PersistTimeOfDay
-getGetter PG.Timestamp = convertPV PersistUTCTime
-getGetter PG.Bit       = convertPV PersistInt64
-getGetter PG.Varbit    = convertPV PersistInt64
-getGetter PG.Numeric   = convertPV (PersistDouble . fromRational)
-getGetter PG.Void      = \_ _ -> Ok PersistNull
+getGetter PG.Bool                  = convertPV PersistBool
+getGetter PG.ByteA                 = convertPV (PersistByteString . unBinary)
+getGetter PG.Char                  = convertPV PersistText
+getGetter PG.Name                  = convertPV PersistText
+getGetter PG.Int8                  = convertPV PersistInt64
+getGetter PG.Int2                  = convertPV PersistInt64
+getGetter PG.Int4                  = convertPV PersistInt64
+getGetter PG.Text                  = convertPV PersistText
+getGetter PG.Xml                   = convertPV PersistText
+getGetter PG.Float4                = convertPV PersistDouble
+getGetter PG.Float8                = convertPV PersistDouble
+getGetter PG.AbsTime               = convertPV PersistUTCTime
+getGetter PG.RelTime               = convertPV PersistUTCTime
+getGetter PG.Money                 = convertPV PersistDouble
+getGetter PG.BpChar                = convertPV PersistText
+getGetter PG.VarChar               = convertPV PersistText
+getGetter PG.Date                  = convertPV PersistDay
+getGetter PG.Time                  = convertPV PersistTimeOfDay
+getGetter PG.Timestamp             = convertPV (PersistUTCTime . localTimeToUTC utc)
+getGetter PG.TimestampTZ           = convertPV (PersistZonedTime . ZT)
+getGetter PG.Bit                   = convertPV PersistInt64
+getGetter PG.VarBit                = convertPV PersistInt64
+getGetter PG.Numeric               = convertPV (PersistDouble . fromRational)
+getGetter PG.Void                  = \_ _ -> Ok PersistNull
 getGetter other   = error $ "Postgresql.getGetter: type " ++
                             show other ++ " not supported."
 
@@ -267,7 +280,7 @@ migrate' :: PersistEntity val
          -> (Text -> IO Statement)
          -> val
          -> IO (Either [Text] [(Bool, Text)])
-migrate' allDefs getter val = do
+migrate' allDefs getter val = fmap (fmap $ map showAlterDb) $ do
     let name = entityDB $ entityDef val
     old <- getColumns getter $ entityDef val
     case partitionEithers old of
@@ -277,7 +290,8 @@ migrate' allDefs getter val = do
             if null old
                 then do
                     let addTable = AddTable $ concat
-                            [ "CREATE TABLE "
+                            -- Lower case e: see Database.Persistent.GenericSql.Migration
+                            [ "CREATe TABLE "
                             , T.unpack $ escape name
                             , "("
                             , T.unpack $ escape $ entityID $ entityDef val
@@ -285,14 +299,15 @@ migrate' allDefs getter val = do
                             , concatMap (\x -> ',' : showColumn x) $ fst new
                             , ")"
                             ]
-                    let rest = flip concatMap (snd new) $ \(uname, ucols) ->
+                    let uniques = flip concatMap (snd new) $ \(uname, ucols) ->
                             [AlterTable name $ AddUniqueConstraint uname ucols]
-                    return $ Right $ map showAlterDb $ addTable : rest
+                        references = mapMaybe (getAddReference name) $ fst new
+                    return $ Right $ addTable : uniques ++ references
                 else do
                     let (acs, ats) = getAlters new old'
                     let acs' = map (AlterColumn name) acs
                     let ats' = map (AlterTable name) ats
-                    return $ Right $ map showAlterDb $ acs' ++ ats'
+                    return $ Right $ acs' ++ ats'
         (errs, _) -> return $ Left errs
 
 data AlterColumn = Type SqlType | IsNull | NotNull | Add Column | Drop
@@ -317,10 +332,10 @@ getColumns getter def = do
             [ PersistText $ unDBName $ entityDB def
             , PersistText $ unDBName $ entityID def
             ]
-    cs <- C.runResourceT $ withStmt stmt vals C.$$ helper
+    cs <- runResourceT $ withStmt stmt vals $$ helper
     stmt' <- getter
         "SELECT constraint_name, column_name FROM information_schema.constraint_column_usage WHERE table_name=? AND column_name <> ? ORDER BY constraint_name, column_name"
-    us <- C.runResourceT $ withStmt stmt' vals C.$$ helperU
+    us <- runResourceT $ withStmt stmt' vals $$ helperU
     return $ cs ++ us
   where
     getAll front = do
@@ -360,18 +375,21 @@ getAlters (c1, u1) (c2, u2) =
     getAltersU :: [(DBName, [DBName])]
                -> [(DBName, [DBName])]
                -> [AlterTable]
-    getAltersU [] old = map (DropConstraint . fst) old
+    getAltersU [] old = map DropConstraint $ filter (not . isManual) $ map fst old
     getAltersU ((name, cols):news) old =
         case lookup name old of
             Nothing -> AddUniqueConstraint name cols : getAltersU news old
             Just ocols ->
                 let old' = filter (\(x, _) -> x /= name) old
-                 in if sort cols == ocols
+                 in if sort cols == sort ocols
                         then getAltersU news old'
                         else  DropConstraint name
                             : AddUniqueConstraint name cols
                             : getAltersU news old'
 
+    -- Don't drop constraints which were manually added.
+    isManual (DBName x) = "__manual_" `T.isPrefixOf` x
+
 getColumn :: (Text -> IO Statement)
           -> DBName -> [PersistValue]
           -> IO (Either Text Column)
@@ -396,27 +414,28 @@ getColumn getter tname [PersistText x, PersistText y, PersistText z, d] =
                 ]
         let ref = refName tname cname
         stmt <- getter sql
-        C.runResourceT $ withStmt stmt
+        runResourceT $ withStmt stmt
                      [ PersistText $ unDBName tname
                      , PersistText $ unDBName ref
-                     ] C.$$ do
+                     ] $$ do
             Just [PersistInt64 i] <- CL.head
             return $ if i == 0 then Nothing else Just (DBName "", ref)
     d' = case d of
             PersistNull   -> Right Nothing
             PersistText t -> Right $ Just t
             _ -> Left $ pack $ "Invalid default column: " ++ show d
-    getType "int4"      = Right $ SqlInt32
-    getType "int8"      = Right $ SqlInteger
-    getType "varchar"   = Right $ SqlString
-    getType "date"      = Right $ SqlDay
-    getType "bool"      = Right $ SqlBool
-    getType "timestamp" = Right $ SqlDayTime
-    getType "float4"    = Right $ SqlReal
-    getType "float8"    = Right $ SqlReal
-    getType "bytea"     = Right $ SqlBlob
-    getType "time"      = Right $ SqlTime
-    getType a           = Left $ "Unknown type: " `T.append` a
+    getType "int4"        = Right $ SqlInt32
+    getType "int8"        = Right $ SqlInt64
+    getType "varchar"     = Right $ SqlString
+    getType "date"        = Right $ SqlDay
+    getType "bool"        = Right $ SqlBool
+    getType "timestamp"   = Right $ SqlDayTime
+    getType "timestamptz" = Right $ SqlDayTimeZoned
+    getType "float4"      = Right $ SqlReal
+    getType "float8"      = Right $ SqlReal
+    getType "bytea"       = Right $ SqlBlob
+    getType "time"        = Right $ SqlTime
+    getType a             = Right $ SqlOther a
 getColumn _ _ x =
     return $ Left $ pack $ "Invalid result from information_schema: " ++ show x
 
@@ -451,6 +470,13 @@ findAlters col@(Column name isNull type_ def _maxLen ref) cols =
              in (modRef ++ modDef ++ modNull ++ modType,
                  filter (\c -> cName c /= name) cols)
 
+-- | Get the references to be added to a table for the given column.
+getAddReference :: DBName -> Column -> Maybe AlterDB
+getAddReference table (Column n _nu _t _def _maxLen ref) =
+    case ref of
+        Nothing -> Nothing
+        Just (s, _) -> Just $ AlterColumn table (n, AddReference s)
+
 showColumn :: Column -> String
 showColumn (Column n nu t def _maxLen ref) = concat
     [ T.unpack $ escape n
@@ -461,21 +487,20 @@ showColumn (Column n nu t def _maxLen ref) = concat
     , case def of
         Nothing -> ""
         Just s -> " DEFAULT " ++ T.unpack s
-    , case ref of
-        Nothing -> ""
-        Just (s, _) -> " REFERENCES " ++ T.unpack (escape s)
     ]
 
 showSqlType :: SqlType -> String
 showSqlType SqlString = "VARCHAR"
 showSqlType SqlInt32 = "INT4"
-showSqlType SqlInteger = "INT8"
+showSqlType SqlInt64 = "INT8"
 showSqlType SqlReal = "DOUBLE PRECISION"
 showSqlType SqlDay = "DATE"
 showSqlType SqlTime = "TIME"
 showSqlType SqlDayTime = "TIMESTAMP"
+showSqlType SqlDayTimeZoned = "TIMESTAMP WITH TIME ZONE"
 showSqlType SqlBlob = "BYTEA"
 showSqlType SqlBool = "BOOLEAN"
+showSqlType (SqlOther t) = T.unpack t
 
 showAlterDb :: AlterDB -> (Bool, Text)
 showAlterDb (AddTable s) = (False, pack s)
@@ -613,7 +638,7 @@ instance PersistConfig PostgresConf where
     loadConfig (Object o) = do
         database <- o .: "database"
         host     <- o .: "host"
-        port     <- o .: "port"
+        port     <- o .:? "port" .!= 5432
         user     <- o .: "user"
         password <- o .: "password"
         pool     <- o .: "poolsize"
